# Contexts
from dataclasses import dataclass, field, replace

import enum
import itertools
from typing import (
    AbstractSet,
    ClassVar,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    cast,
)
from yarl import URL

from . import ast
from .expr import LiteralT, RootABC, TypeT
from .types import LocalPath, RemotePath


# Neuro-flow contexts (variables available during expressions calculation).

# The basic design idea is: we should avoid nested contexts if possible.
#
# That's why there is `env` context but defaults.env and job.env are absent.
#
# During the calculation the `env` context is updated to reflect global envs and job's
# env accordingly. This design principle makes expressions shorter in yaml file.  Short
# expressions are super important since the expression syntax has no user-defined
# variables.


# neuro -- global settings (cluster, user, api entrypoint)

# flow -- global flow settings, e.g. id

# defaults -- global flow defaults: env, preset, tags, workdir, life_span

# env -- Contains environment variables set in a workflow, job, or step

# job -- Information about the currently executing job

# batch -- Information about the currently executing batch

# steps -- Information about the steps that have been run in this job

# secrets -- Enables access to secrets.
#
# Do we want the explicit secrets support?  Perhaps better to have the secrets
# substitution on the client's cluster side (even not on the platform-api).

# strategy -- Enables access to the configured strategy parameters.
#
# Strategy parameters include fail-fast and max-parallel.

# matrix -- Enables access to the matrix parameters you configured for the current job.

# needs -- Enables access to the outputs of all jobs that are defined as a dependency of
# the current job.

# images -- Enables access to image specifications.

# volumes -- Enables access to volume specifications.


class NotAvailable(LookupError):
    def __init__(self, ctx_name: str) -> None:
        super().__init__(f"Context {ctx_name} is not available")


class UnknownJob(KeyError):
    pass


class UnknownTask(KeyError):
    pass


@dataclass(frozen=True)
class Neuro:
    pass


class Result(str, enum.Enum):
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELLED = "cancelled"


@dataclass(frozen=True)
class DepCtx:
    result: Result
    outputs: Mapping[str, str]


NeedsCtx = Mapping[str, DepCtx]


MatrixCtx = Mapping[str, LiteralT]


@dataclass(frozen=True)
class StrategyCtx:
    fail_fast: bool = False
    max_parallel: int = 10


@dataclass(frozen=True)
class ExecUnitCtx:
    title: Optional[str]
    name: Optional[str]
    image: str
    preset: Optional[str]
    http_port: Optional[int]
    http_auth: Optional[bool]
    entrypoint: Optional[str]
    cmd: Optional[str]
    workdir: Optional[RemotePath]
    volumes: Sequence[str]  # Sequence[VolumeRef]
    tags: AbstractSet[str]
    life_span: Optional[float]


@dataclass(frozen=True)
class JobCtx(ExecUnitCtx):
    id: str
    detach: bool
    browse: bool
    port_forward: Sequence[str]


@dataclass(frozen=True)
class PreparedTaskCtx:
    id: Optional[str]
    real_id: str

    needs: AbstractSet[str]  # A set of batch.id
    matrix: MatrixCtx
    strategy: StrategyCtx

    ast: ast.Task


@dataclass(frozen=True)
class TaskCtx(ExecUnitCtx):
    # executed task
    id: Optional[str]
    real_id: str

    needs: AbstractSet[str]  # A set of batch.id

    # continue_on_error: OptBoolExpr
    # if_: OptBoolExpr  # -- skip conditionally


@dataclass(frozen=True)
class VolumeCtx:
    id: str
    remote: URL
    mount: RemotePath
    read_only: bool
    local: Optional[LocalPath]
    full_local_path: Optional[LocalPath]

    @property
    def ref_ro(self) -> str:
        return f"{self.remote}:{self.mount}:ro"

    @property
    def ref_rw(self) -> str:
        return f"{self.remote}:{self.mount}:rw"

    @property
    def ref(self) -> str:
        ro = "ro" if self.read_only else "rw"
        return f"{self.remote}:{self.mount}:{ro}"


@dataclass(frozen=True)
class ImageCtx:
    id: str
    ref: str
    context: Optional[LocalPath]
    full_context_path: Optional[LocalPath]
    dockerfile: Optional[LocalPath]
    full_dockerfile_path: Optional[LocalPath]
    build_args: Sequence[str]


@dataclass(frozen=True)
class DefaultsCtx:
    tags: AbstractSet[str]
    workdir: Optional[RemotePath]
    life_span: Optional[float]
    preset: Optional[str]


# @dataclass(frozen=True)
# class BatchDefaultsCtx(DefaultsCtx):
#     fail_fast: Optional[bool]
#     max_parallel: Optional[int]


@dataclass(frozen=True)
class FlowCtx:
    id: str
    workspace: LocalPath
    title: str


_CtxT = TypeVar("_CtxT", bound="BaseContext")


@dataclass(frozen=True)
class BaseContext(RootABC):
    FLOW_TYPE: ClassVar[Type[ast.BaseFlow]] = field(init=False)
    LOOKUP_KEYS: ClassVar[Tuple[str, ...]] = field(
        init=False,
        default=("flow", "defaults", "volumes", "images", "env", "job", "batch",),
    )

    _ast_flow: ast.BaseFlow
    flow: FlowCtx
    _defaults: Optional[DefaultsCtx] = None
    _env: Optional[Mapping[str, str]] = None

    _images: Optional[Mapping[str, ImageCtx]] = None
    _volumes: Optional[Mapping[str, VolumeCtx]] = None

    # Add a context with global flow info, e.g. ctx.flow.id maybe?

    @classmethod
    async def create(cls: Type[_CtxT], ast_flow: ast.BaseFlow) -> _CtxT:
        assert isinstance(ast_flow, cls.FLOW_TYPE)
        flow = FlowCtx(
            id=ast_flow.id,
            workspace=ast_flow.workspace.resolve(),
            title=ast_flow.title or ast_flow.id,
        )

        ctx = cls(_ast_flow=ast_flow, flow=flow,)

        ast_defaults = ast_flow.defaults
        if ast_defaults is not None:
            if ast_defaults.env is not None:
                env = {k: await v.eval(ctx) for k, v in ast_defaults.env.items()}
            else:
                env = {}

            if ast_defaults.tags is not None:
                tags = {await t.eval(ctx) for t in ast_defaults.tags}
            else:
                tags = set()
            workdir = await ast_defaults.workdir.eval(ctx)
            life_span = await ast_defaults.life_span.eval(ctx)
            preset = await ast_defaults.preset.eval(ctx)
        else:
            env = {}
            tags = set()
            workdir = None
            life_span = None
            preset = None

        if not tags:
            tags = {f"flow:{flow.id}"}

        defaults = DefaultsCtx(
            tags=tags, workdir=workdir, life_span=life_span, preset=preset,
        )
        ctx = replace(ctx, _defaults=defaults, _env=env)

        # volumes / images needs a context with defaults only for self initialization
        volumes = {}
        if ast_flow.volumes is not None:
            for k, v in ast_flow.volumes.items():
                local_path = await v.local.eval(ctx)
                volumes[k] = VolumeCtx(
                    id=k,
                    remote=await v.remote.eval(ctx),
                    mount=await v.mount.eval(ctx),
                    read_only=bool(await v.read_only.eval(ctx)),
                    local=local_path,
                    full_local_path=calc_full_path(ctx, local_path),
                )
        images = {}
        if ast_flow.images is not None:
            for k, i in ast_flow.images.items():
                context_path = await i.context.eval(ctx)
                dockerfile_path = await i.dockerfile.eval(ctx)
                if i.build_args is not None:
                    build_args = [await v.eval(ctx) for v in i.build_args]
                else:
                    build_args = []
                images[k] = ImageCtx(
                    id=k,
                    ref=await i.ref.eval(ctx),
                    context=context_path,
                    full_context_path=calc_full_path(ctx, context_path),
                    dockerfile=dockerfile_path,
                    full_dockerfile_path=calc_full_path(ctx, dockerfile_path),
                    build_args=build_args,
                )
        return replace(ctx, _volumes=volumes, _images=images)

    def lookup(self, name: str) -> TypeT:
        if name not in self.LOOKUP_KEYS:
            raise NotAvailable(name)
        ret = getattr(self, name)
        # assert isinstance(ret, (ContainerT, SequenceT, MappingT)), ret
        return cast(TypeT, ret)

    @property
    def env(self) -> Mapping[str, str]:
        if self._env is None:
            raise NotAvailable("env")
        return self._env

    @property
    def defaults(self) -> DefaultsCtx:
        if self._defaults is None:
            raise NotAvailable("defaults")
        return self._defaults

    @property
    def volumes(self) -> Mapping[str, VolumeCtx]:
        if self._volumes is None:
            raise NotAvailable("volumes")
        return self._volumes

    @property
    def images(self) -> Mapping[str, ImageCtx]:
        if self._images is None:
            raise NotAvailable("images")
        return self._images


@dataclass(frozen=True)
class LiveContext(BaseContext):
    FLOW_TYPE: ClassVar[Type[ast.LiveFlow]] = field(init=False, default=ast.LiveFlow)
    LOOKUP_KEYS: ClassVar[Tuple[str, ...]] = field(
        init=False, default=BaseContext.LOOKUP_KEYS + ("job",)
    )
    _job: Optional[JobCtx] = None

    @property
    def job(self) -> JobCtx:
        if self._job is None:
            raise NotAvailable("job")
        return self._job

    async def with_job(self, job_id: str) -> "LiveContext":
        if self._job is not None:
            raise TypeError(
                "Cannot enter into the job context, if job is already initialized"
            )
        assert isinstance(self._ast_flow, self.FLOW_TYPE)

        try:
            job = self._ast_flow.jobs[job_id]
        except KeyError:
            raise UnknownJob(job_id)

        tags = set()
        if job.tags is not None:
            tags = {await v.eval(self) for v in job.tags}
        if not tags:
            tags = {f"job:{job_id}"}

        env = dict(self.env)
        if job.env is not None:
            env.update({k: await v.eval(self) for k, v in job.env.items()})

        title = await job.title.eval(self)
        if title is None:
            title = f"{self.flow.id}.{job_id}"

        workdir = (await job.workdir.eval(self)) or self.defaults.workdir

        volumes = []
        if job.volumes is not None:
            volumes = [await v.eval(self) for v in job.volumes]

        life_span = (await job.life_span.eval(self)) or self.defaults.life_span

        preset = (await job.preset.eval(self)) or self.defaults.preset
        port_forward = []
        if job.port_forward is not None:
            port_forward = [await val.eval(self) for val in job.port_forward]

        job_ctx = JobCtx(
            id=job_id,
            detach=bool(await job.detach.eval(self)),
            browse=bool(await job.browse.eval(self)),
            title=title,
            name=(await job.name.eval(self)),
            image=await job.image.eval(self),
            preset=preset,
            entrypoint=await job.entrypoint.eval(self),
            cmd=await job.cmd.eval(self),
            workdir=workdir,
            volumes=volumes,
            tags=self.defaults.tags | tags,
            life_span=life_span,
            http_port=await job.http_port.eval(self),
            http_auth=await job.http_auth.eval(self),
            port_forward=port_forward,
        )
        return replace(self, _job=job_ctx, _env=env,)


@dataclass(frozen=True)
class BatchContext(BaseContext):
    FLOW_TYPE: ClassVar[Type[ast.BatchFlow]] = field(init=False, default=ast.BatchFlow)
    LOOKUP_KEYS: ClassVar[Tuple[str, ...]] = field(
        init=False,
        default=BaseContext.LOOKUP_KEYS + ("batch", "needs", "matrix", "strategy"),
    )
    _task: Optional[TaskCtx] = None
    _needs: Optional[NeedsCtx] = None
    _prep_tasks: Optional[Mapping[str, PreparedTaskCtx]] = None
    _graph: Optional[Mapping[str, AbstractSet[str]]] = None
    _matrix: Optional[MatrixCtx] = None
    _strategy: Optional[StrategyCtx] = None

    @classmethod
    async def create(cls: Type[_CtxT], ast_flow: ast.BaseFlow) -> _CtxT:
        ctx = await super(cls, BatchContext).create(ast_flow)
        assert isinstance(ctx._ast_flow, ast.BatchFlow)
        prep_tasks = {}
        last_needs: Set[str] = set()
        for num, ast_task in enumerate(ctx._ast_flow.tasks, 1):
            # eval matrix
            matrix: Sequence[MatrixCtx]
            strategy: StrategyCtx
            default_strategy = StrategyCtx()
            if ast_task.strategy is not None:
                fail_fast = await ast_task.strategy.fail_fast.eval(ctx)
                if fail_fast is None:
                    fail_fast = default_strategy.fail_fast
                max_parallel = await ast_task.strategy.max_parallel.eval(ctx)
                if max_parallel is None:
                    max_parallel = default_strategy.max_parallel
                strategy = StrategyCtx(fail_fast=fail_fast, max_parallel=max_parallel,)
                if ast_task.strategy.matrix is not None:
                    matrix = await ctx._build_matrix(ast_task.strategy)
                    matrix = await ctx._exclude(ast_task.strategy, matrix)
                    matrix = await ctx._include(ast_task.strategy, matrix)
                else:
                    matrix = [{}]  # dummy
            else:
                strategy = default_strategy  # default
                matrix = [{}]  # dummy

            if len(matrix) > 256:
                raise ValueError(
                    f"The matrix size for task #{num} exceeds the limit of 256"
                )

            real_ids = set()
            for row in matrix:
                # make prep patch(es)
                matrix_ctx = await ctx.with_matrix(row)
                task_id = await ast_task.id.eval(matrix_ctx)
                if task_id is None:
                    # Dash is not allowed in identifier, so the generated read id
                    # never clamps with user_provided one.
                    suffix = [str(row[k]) for k in sorted(row)]
                    real_id = "-".join(["batch", str(num), *suffix])
                else:
                    real_id = task_id
                if real_id in prep_tasks:
                    raise ValueError(f"Duplicated task id {real_id}")
                if ast_task.needs is not None:
                    needs = {await need.eval(matrix_ctx) for need in ast_task.needs}
                else:
                    needs = last_needs

                prep = PreparedTaskCtx(
                    id=task_id,
                    real_id=real_id,
                    needs=needs,
                    matrix=row,
                    strategy=strategy,
                    ast=ast_task,
                )
                prep_tasks[real_id] = prep
                real_ids.add(real_id)

            last_needs = real_ids

        graph: Dict[str, AbstractSet[str]] = {}
        for key, val in prep_tasks.items():
            graph[key] = val.needs

        return replace(  # type: ignore[return-value]
            ctx, _prep_tasks=prep_tasks, _graph=graph
        )

    @property
    def task(self) -> TaskCtx:
        if self._task is None:
            raise NotAvailable("task")
        return self._task

    @property
    def needs(self) -> NeedsCtx:
        if self._needs is None:
            raise NotAvailable("needs")
        return self._needs

    @property
    def matrix(self) -> MatrixCtx:
        if self._matrix is None:
            raise NotAvailable("matrix")
        return self._matrix

    @property
    def strategy(self) -> StrategyCtx:
        if self._strategy is None:
            raise NotAvailable("strategy")
        return self._strategy

    @property
    def graph(self) -> Mapping[str, AbstractSet[str]]:
        # Batch names, sorted by the execution order.
        # Batches from each set in the list can be executed concurrently.
        assert self._graph is not None
        return self._graph

    @property
    def cardinality(self) -> int:
        # 1 slot for started task and another one for finished
        assert self._prep_tasks is not None
        return len(self._prep_tasks) * 2

    def get_dep_ids(self, real_id: str) -> AbstractSet[str]:
        assert self._prep_tasks is not None
        prep_task = self._prep_tasks[real_id]
        return prep_task.needs

    async def with_matrix(self, matrix: MatrixCtx) -> "BatchContext":
        if self._matrix is not None:
            raise TypeError(
                "Cannot enter into the matrix context if "
                "the matrix is already initialized"
            )
        return replace(self, _matrix=matrix)

    async def with_task(self, real_id: str, *, needs: NeedsCtx) -> "BatchContext":
        # real_id -- the task's real id
        #
        # outputs -- real_id -> (output_name -> value) mapping for all task ids
        # enumerated in needs.
        #
        # TODO: multi-state batches require 'state' mapping (state_name -> value)
        assert self._prep_tasks is not None

        if self._task is not None:
            raise TypeError(
                "Cannot enter into the task context if "
                "the task is already initialized"
            )
        try:
            prep_task = self._prep_tasks[real_id]
        except KeyError:
            raise UnknownTask(real_id)

        if needs.keys() != prep_task.needs:
            extra = ",".join(needs.keys() - prep_task.needs)
            missing = ",".join(prep_task.needs - needs.keys())
            err = ["Error in 'needs':"]
            if extra:
                err.append(f"unexpected keys {extra}")
            if missing:
                err.append(f"missing keys {missing}")
            raise ValueError(" ".join(err))

        ctx = await self.with_matrix(prep_task.matrix)

        env = dict(ctx.env)
        if prep_task.ast.env is not None:
            env.update({k: await v.eval(ctx) for k, v in prep_task.ast.env.items()})

        title = await prep_task.ast.title.eval(ctx)
        if title is None:
            title = f"{ctx.flow.id}.{real_id}"
        title = await prep_task.ast.title.eval(ctx)

        tags = set()
        if prep_task.ast.tags is not None:
            tags = {await v.eval(ctx) for v in prep_task.ast.tags}
        if not tags:
            tags = {f"batch:{real_id}"}

        workdir = (await prep_task.ast.workdir.eval(ctx)) or ctx.defaults.workdir

        volumes = []
        if prep_task.ast.volumes is not None:
            volumes = [await v.eval(ctx) for v in prep_task.ast.volumes]

        life_span = (await prep_task.ast.life_span.eval(ctx)) or ctx.defaults.life_span

        preset = (await prep_task.ast.preset.eval(ctx)) or ctx.defaults.preset

        task_ctx = TaskCtx(
            id=prep_task.id,
            real_id=prep_task.real_id,
            needs=prep_task.needs,
            title=title,
            name=(await prep_task.ast.name.eval(ctx)),
            image=await prep_task.ast.image.eval(ctx),
            preset=preset,
            entrypoint=await prep_task.ast.entrypoint.eval(ctx),
            cmd=await prep_task.ast.cmd.eval(ctx),
            workdir=workdir,
            volumes=volumes,
            tags=ctx.defaults.tags | tags,
            life_span=life_span,
            http_port=await prep_task.ast.http_port.eval(ctx),
            http_auth=await prep_task.ast.http_auth.eval(ctx),
        )
        return replace(
            ctx, _task=task_ctx, _env=env, _needs=needs, _strategy=prep_task.strategy
        )

    async def _build_matrix(self, strategy: ast.Strategy) -> Sequence[MatrixCtx]:
        assert strategy.matrix is not None
        products = []
        for k, lst in strategy.matrix.products.items():
            lst2 = [{k: await i.eval(self)} for i in lst]
            products.append(lst2)
        ret = []
        for row in itertools.product(*products):
            dct: Dict[str, LiteralT] = {}
            for elem in row:
                dct.update(elem)
            ret.append(dct)
        return ret

    async def _exclude(
        self, strategy: ast.Strategy, matrix: Sequence[MatrixCtx]
    ) -> Sequence[MatrixCtx]:
        assert strategy.matrix is not None
        exclude = []
        for dct in strategy.matrix.exclude:
            exclude.append({k: await v.eval(self) for k, v in dct.items()})
        ret = []
        for row in matrix:
            for exc in exclude:
                match = True
                for k, v in exc.items():
                    if row[k] != v:
                        match = False
                        break
                if not match:
                    ret.append(row)
        return ret

    async def _include(
        self, strategy: ast.Strategy, matrix: Sequence[MatrixCtx]
    ) -> Sequence[MatrixCtx]:
        assert strategy.matrix is not None
        ret = list(matrix)
        for dct in strategy.matrix.include:
            ret.append({k: await v.eval(self) for k, v in dct.items()})
        return ret


def calc_full_path(ctx: BaseContext, path: Optional[LocalPath]) -> Optional[LocalPath]:
    if path is None:
        return None
    if path.is_absolute():
        return path
    return ctx.flow.workspace.joinpath(path).resolve()



# Dataclasses
from dataclasses import dataclass

import enum
from typing import Mapping, Optional, Sequence

from .expr import (
    IdExpr,
    OptBoolExpr,
    OptIdExpr,
    OptIntExpr,
    OptLifeSpanExpr,
    OptLocalPathExpr,
    OptRemotePathExpr,
    OptStrExpr,
    PortPairExpr,
    RemotePathExpr,
    StrExpr,
    URIExpr,
)
from .tokenizer import Pos
from .types import LocalPath


@dataclass(frozen=True)
class Base:
    _start: Pos
    _end: Pos


# There are 'batch' for pipelined mode and 'live' for interactive one
# (while 'batches' are technically just non-interactive jobs.


class Kind(enum.Enum):
    LIVE = "live"  # interactive mode.
    BATCH = "batch"  # pipelined mode


@dataclass(frozen=True)
class Volume(Base):
    remote: URIExpr  # remote URI, e.g. storage:folder/subfolder
    mount: RemotePathExpr  # mount path inside container
    local: OptLocalPathExpr
    read_only: OptBoolExpr  # True if mounted in read-only mode, False for read-write


@dataclass(frozen=True)
class Image(Base):
    ref: StrExpr  # Image reference, e.g. image:my-proj or neuromation/base@v1.6
    context: OptLocalPathExpr
    dockerfile: OptLocalPathExpr
    build_args: Optional[Sequence[StrExpr]]


@dataclass(frozen=True)
class ExecUnit(Base):
    title: OptStrExpr  # Autocalculated if not passed explicitly
    name: OptStrExpr
    image: StrExpr
    preset: OptStrExpr
    entrypoint: OptStrExpr
    cmd: OptStrExpr
    workdir: OptRemotePathExpr
    env: Optional[Mapping[str, StrExpr]]
    volumes: Optional[Sequence[StrExpr]]
    tags: Optional[Sequence[StrExpr]]
    life_span: OptLifeSpanExpr
    http_port: OptIntExpr
    http_auth: OptBoolExpr


@dataclass(frozen=True)
class Matrix(Base):
    # AST class is slightly different from YAML representation,
    # in YAML `products` mapping is embedded into the matrix itself.
    products: Mapping[str, Sequence[StrExpr]]
    exclude: Sequence[Mapping[str, StrExpr]]
    include: Sequence[Mapping[str, StrExpr]]


@dataclass(frozen=True)
class Strategy(Base):
    matrix: Matrix
    fail_fast: OptBoolExpr
    max_parallel: OptIntExpr


@dataclass(frozen=True)
class Job(ExecUnit):
    # Interactive job used by Kind.Live flow

    detach: OptBoolExpr
    browse: OptBoolExpr
    port_forward: Optional[Sequence[PortPairExpr]]


@dataclass(frozen=True)
class Task(ExecUnit):
    id: OptIdExpr

    # A set of steps, used in net mode
    # All steps share the same implicit persistent disk volume

    needs: Optional[Sequence[IdExpr]]  # BatchRef

    # matrix? Do we need a build matrix? Yes probably.

    strategy: Optional[Strategy]

    # continue_on_error: OptBoolExpr
    # if_: OptBoolExpr  # -- skip conditionally


@dataclass(frozen=True)
class FlowDefaults(Base):
    tags: Optional[Sequence[StrExpr]]

    env: Optional[Mapping[str, StrExpr]]
    workdir: OptRemotePathExpr

    life_span: OptLifeSpanExpr

    preset: OptStrExpr


# @dataclass(frozen=True)
# class BatchFlowDefaults(FlowDefaults):
#     fail_fast: OptBoolExpr
#     max_parallel: OptIntExpr


@dataclass(frozen=True)
class BaseFlow(Base):
    kind: Kind
    id: str
    workspace: LocalPath

    title: Optional[str]

    images: Optional[Mapping[str, Image]]
    volumes: Optional[Mapping[str, Volume]]
    defaults: Optional[FlowDefaults]


@dataclass(frozen=True)
class LiveFlow(BaseFlow):
    # self.kind == Kind.Job
    jobs: Mapping[str, Job]


@dataclass(frozen=True)
class BatchFlow(BaseFlow):
    # self.kind == Kind.Batch
    tasks: Sequence[Task]



import click
import functools
import inspect
import logging
import sys
from neuromation.api import get as api_get
from neuromation.cli.asyncio_utils import Runner
from typing import Any, Awaitable, Callable, Optional, TypeVar

from .batch_runner import BatchRunner
from .live_runner import LiveRunner
from .parser import ConfigDir, find_live_config, find_workspace, parse_live
from .storage import BatchFSStorage, BatchStorage


if sys.version_info >= (3, 7):
    from contextlib import AsyncExitStack
else:
    from async_exit_stack import AsyncExitStack


_T = TypeVar("_T")


def wrap_async(callback: Callable[..., Awaitable[_T]],) -> Callable[..., _T]:
    assert inspect.iscoroutinefunction(callback)

    # N.B. the decorator implies @click.pass_obj
    @click.pass_obj
    @functools.wraps(callback)
    def wrapper(*args: Any, **kwargs: Any) -> _T:
        with Runner() as runner:
            return runner.run(callback(*args, **kwargs))

    return wrapper


@click.group()
@click.option(
    "--config",
    type=click.Path(dir_okay=True, file_okay=False),
    required=False,
    help=(
        "Path to a directory with .neuro folder inside, "
        "automatic lookup is performed if not set (default)"
    ),
    default=None,
    metavar="PATH",
)
@click.pass_context
def main(ctx: click.Context, config: Optional[str]) -> None:
    logging.basicConfig(level=logging.INFO)
    config_dir = find_workspace(config)
    ctx.obj = config_dir


# #### job commands ####


@main.command()
@wrap_async
async def ps(config_dir: ConfigDir) -> None:
    """List all jobs"""
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        await runner.ps()


@main.command()
@click.argument("job-id")
@wrap_async
async def run(config_dir: ConfigDir, job_id: str) -> None:
    """Run a job.

    RUN job JOB-ID or ATTACH to it if the job is already running
    """
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        await runner.run(job_id)


@main.command()
@click.argument("job-id")
@wrap_async
async def logs(config_dir: ConfigDir, job_id: str) -> None:
    """Print logs.

    Displys logs for JOB-ID
    """
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        await runner.logs(job_id)


@main.command()
@click.argument("job-id")
@wrap_async
async def status(config_dir: ConfigDir, job_id: str) -> None:
    """Show job status.

    Print status for JOB-ID
    """
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        await runner.status(job_id)


@main.command()
@click.argument("job-id")
@wrap_async
async def kill(config_dir: ConfigDir, job_id: str) -> None:
    """Kill a job.

    Kill JOB-ID, use `kill ALL` for killing all jobs."""
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        if job_id != "ALL":
            await runner.kill(job_id)
        else:
            await runner.kill_all()


# #### storage commands ####


@main.command()
@click.argument("volume")
@wrap_async
async def upload(config_dir: ConfigDir, volume: str) -> None:
    """Upload volume.

    Upload local files to remote for VOLUME,
    use `upload ALL` for uploading all volumes."""
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        if volume != "ALL":
            await runner.upload(volume)
        else:
            await runner.upload_all()


@main.command()
@click.argument("volume")
@wrap_async
async def download(config_dir: ConfigDir, volume: str) -> None:
    """Download volume.

    Download remote files to local for VOLUME,
    use `download ALL` for downloading all volumes."""
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        if volume != "ALL":
            await runner.download(volume)
        else:
            await runner.download_all()


@main.command()
@click.argument("volume")
@wrap_async
async def clean(config_dir: ConfigDir, volume: str) -> None:
    """Clean volume.

    Clean remote files on VOLUME,
    use `clean ALL` for cleaning up all volumes."""
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        if volume != "ALL":
            await runner.clean(volume)
        else:
            await runner.clean_all()


@main.command()
@wrap_async
async def mkvolumes(config_dir: ConfigDir) -> None:
    """Create all remote folders for volumes."""
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        await runner.mkvolumes()


# #### image commands ####


@main.command()
@click.argument("image")
@wrap_async
async def build(config_dir: ConfigDir, image: str) -> None:
    """Build an image.

    Assemble the IMAGE remotely and publish it.
    """
    config_path = find_live_config(config_dir)
    flow = parse_live(config_path.workspace, config_path.config_file)
    async with LiveRunner(flow) as runner:
        await runner.build(image)


# #### pipeline commands ####


@main.command()
@click.argument("batch")
@wrap_async
async def start(config_dir: ConfigDir, batch: str) -> None:
    """Start a pipeline.

    Run BATCH remotely on the cluster.
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(api_get())
        storage: BatchStorage = await stack.enter_async_context(BatchFSStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(config_dir, client, storage)
        )
        await runner.bake(batch)



import dataclasses

import abc
import datetime
import json
import re
import secrets
import sys
from neuromation.api import Client, JobDescription, JobStatus
from types import TracebackType
from typing import Any, Dict, Optional, Tuple, Type
from typing_extensions import Final
from yarl import URL

from .context import Result


if sys.version_info < (3, 7):
    from backports.datetime_fromisoformat import MonkeyPatch

    MonkeyPatch.patch_fromisoformat()


STARTED_RE: Final = re.compile(r"\A\d+\.(?P<id>[a-zA-Z][a-zA-Z0-9_\-]*).started.json\Z")
FINISHED_RE: Final = re.compile(
    r"\A\d+\.(?P<id>[a-zA-Z][a-zA-Z0-9_\-]*).finished.json\Z"
)


@dataclasses.dataclass(frozen=True)
class BakeInit:
    id: str
    config_file: URL
    cardinality: int
    when: datetime.datetime


@dataclasses.dataclass(frozen=True)
class StartedTask:
    id: str
    raw_id: str
    created_at: datetime.datetime
    when: datetime.datetime


@dataclasses.dataclass(frozen=True)
class FinishedTask:
    id: str
    raw_id: str
    when: datetime.datetime
    status: JobStatus
    exit_code: Optional[int]
    created_at: datetime.datetime
    started_at: datetime.datetime
    finished_at: datetime.datetime
    finish_reason: str
    finish_description: str

    @property
    def result(self) -> Result:
        if self.status == JobStatus.SUCCEEDED:
            return Result.SUCCEEDED
        else:
            return Result.FAILED


# A storage abstraction
#
# There is a possibility to add Postgres storage class later, for example


class BatchStorage(abc.ABC):
    async def __aenter__(self) -> "BatchStorage":
        return self

    async def __aexit__(
        self,
        exc_typ: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    @abc.abstractmethod
    async def close(self) -> None:
        pass

    @abc.abstractmethod
    async def create_bake(
        self, batch_name: str, config_name: str, config_content: str, cardinality: int,
    ) -> str:
        pass

    @abc.abstractmethod
    async def fetch_bake(self, bake_id: str) -> BakeInit:
        pass

    @abc.abstractmethod
    async def fetch_config(self, bake_id: str, config_name: str) -> str:
        pass

    @abc.abstractmethod
    async def create_attempt(
        self, bake_id: str, attempt_no: int, cardinality: int
    ) -> None:
        pass

    @abc.abstractmethod
    async def find_last_attempt(self, bake_id: str) -> int:
        pass

    @abc.abstractmethod
    async def fetch_attempt(
        self, bake_id: str, attempt_no: int, cardinality: int
    ) -> Tuple[Dict[str, FinishedTask], Dict[str, StartedTask]]:
        pass

    @abc.abstractmethod
    async def finish_attempt(
        self, bake_id: str, attempt_no: int, cardinality: int, result: Result
    ) -> None:
        pass

    @abc.abstractmethod
    async def start_task(
        self,
        bake_id: str,
        attempt_no: int,
        task_no: int,
        cardinality: int,
        task_id: str,
        descr: JobDescription,
    ) -> StartedTask:
        pass

    @abc.abstractmethod
    async def finish_task(
        self,
        bake_id: str,
        attempt_no: int,
        task_no: int,
        cardinality: int,
        task: StartedTask,
        descr: JobDescription,
    ) -> FinishedTask:
        pass


class BatchFSStorage(BatchStorage):
    # A storage that uses storage:.flow directory as a database

    # A lack of true atomic operations (and server-side move/rename which prevents
    # overriding the destination) on the storage fs makes the implementation slightly
    # subtle but it's ok.
    #
    # The genuine consistent storage like PostgresSQL-based dedicated Neuro platform
    # service can be added later as a separate implementation that follows the same
    # abstract API.

    # The FS structure:
    # storage:.flow
    # +-- bake_id
    #     +-- config.yml
    #     +-- 00.init.json
    #     +-- 01.attempt
    #         +-- 000.init.json
    #         +-- 001.<task_id>.started.json
    #         +-- 002.<task_id>.finished.json
    #         +-- 999.result.json
    #     +-- 02.attempt
    #         +-- 000.init.json
    #         +-- 001.<task_id>.started.json
    #         +-- 002.<task_id>.finished.json
    #         +-- 999.result.json

    def __init__(self, client: Client) -> None:
        self._client = client

    def _now(self) -> str:
        dt = datetime.datetime.now(datetime.timezone.utc)
        return dt.isoformat(timespec="seconds")

    def _mk_bake_uri(self, bake_id: str) -> URL:
        return URL("storage:.flow") / bake_id

    async def close(self) -> None:
        pass

    async def create_bake(
        self, batch_name: str, config_name: str, config_content: str, cardinality: int
    ) -> str:
        # Return bake_id
        bake_id = "_".join([batch_name, self._now(), secrets.token_hex(3)])
        bake_uri = self._mk_bake_uri(bake_id)
        await self._client.storage.mkdir(bake_uri, parents=True)
        config_uri = bake_uri / config_name
        await self._write_file(config_uri, config_content)
        started = {
            "config_file": config_uri.name,
            "id": batch_name,
            "cardinality": cardinality,
        }
        await self._write_json(bake_uri / "00.init.json", started)
        return bake_id

    async def fetch_bake(self, bake_id: str) -> BakeInit:
        data = await self._read_json(self._mk_bake_uri(bake_id) / "00.init.json")
        return BakeInit(
            id=data["id"],
            config_file=URL(data["config_file"]),
            cardinality=data["cardinality"],
            when=datetime.datetime.fromisoformat(data["when"]),
        )

    async def fetch_config(self, bake_id: str, config_name: str) -> str:
        return await self._read_file(self._mk_bake_uri(bake_id) / config_name)

    async def create_attempt(
        self, bake_id: str, attempt_no: int, cardinality: int
    ) -> None:
        assert 0 < attempt_no < 100, attempt_no
        bake_uri = self._mk_bake_uri(bake_id)
        attempt_uri = bake_uri / f"{attempt_no:02d}.attempt"
        await self._client.storage.mkdir(attempt_uri)
        digits = cardinality // 10 + 1
        pre = "0".zfill(digits)
        await self._write_json(
            attempt_uri / f"{pre}.init.json", {"cardinality": cardinality}
        )

    async def find_last_attempt(self, bake_id: str) -> int:
        bake_uri = self._mk_bake_uri(bake_id)
        files = set()
        async for fi in self._client.storage.ls(bake_uri):
            files.add(fi.name)
        if "00.init.json" not in files:
            raise ValueError("The batch is not initialized properly")
        for attempt_no in range(99, 0, -1):
            if f"{attempt_no:02d}.attempt" in files:
                return attempt_no
        assert False, "unreachable"

    async def fetch_attempt(
        self, bake_id: str, attempt_no: int, cardinality: int
    ) -> Tuple[Dict[str, FinishedTask], Dict[str, StartedTask]]:
        bake_uri = self._mk_bake_uri(bake_id)
        attempt_url = bake_uri / f"{attempt_no:02d}.attempt"
        digits = cardinality // 10 + 1
        pre = "0".zfill(digits)
        init_name = f"{pre}.init.json"
        data = await self._read_json(attempt_url / init_name)
        assert data["cardinality"] == cardinality
        started = {}
        finished = {}
        async for fs in self._client.storage.ls(attempt_url):
            if fs.name == init_name:
                continue
            match = STARTED_RE.match(fs.name)
            if match:
                data = await self._read_json(attempt_url / fs.name)
                assert match.group("id") == data["id"]
                started[data["id"]] = StartedTask(
                    data["id"],
                    data["raw_id"],
                    datetime.datetime.fromisoformat(data["created_at"]),
                    datetime.datetime.fromisoformat(data["when"]),
                )
                continue
            match = FINISHED_RE.match(fs.name)
            if match:
                data = await self._read_json(attempt_url / fs.name)
                assert match.group("id") == data["id"]
                finished[data["id"]] = FinishedTask(
                    data["id"],
                    data["raw_id"],
                    datetime.datetime.fromisoformat(data["when"]),
                    JobStatus(data["status"]),
                    data["exit_code"],
                    datetime.datetime.fromisoformat(data["created_at"]),
                    datetime.datetime.fromisoformat(data["started_at"]),
                    datetime.datetime.fromisoformat(data["finished_at"]),
                    finish_reason=data["finish_reason"],
                    finish_description=data["finish_description"],
                )
                continue
            raise ValueError(f"Unexpected name {attempt_url / fs.name}")
        assert finished.keys() <= started.keys()
        return finished, started

    async def finish_attempt(
        self, bake_id: str, attempt_no: int, cardinality: int, result: Result
    ) -> None:
        bake_uri = self._mk_bake_uri(bake_id)
        attempt_url = bake_uri / f"{attempt_no:02d}.attempt"
        digits = cardinality // 10 + 1
        pre = "9" * digits
        data = {"result": str(result)}
        await self._write_json(attempt_url / f"{pre}.result.json", data)

    async def start_task(
        self,
        bake_id: str,
        attempt_no: int,
        task_no: int,
        cardinality: int,
        task_id: str,
        descr: JobDescription,
    ) -> StartedTask:
        bake_uri = self._mk_bake_uri(bake_id)
        attempt_url = bake_uri / f"{attempt_no:02d}.attempt"
        digits = cardinality // 10 + 1
        pre = str(task_no + 1).zfill(digits)
        assert descr.history.created_at is not None
        ret = StartedTask(
            id=task_id,
            raw_id=descr.id,
            when=datetime.datetime.now(datetime.timezone.utc),
            created_at=descr.history.created_at,
        )

        data = {
            "id": ret.id,
            "raw_id": ret.raw_id,
            "when": ret.when.isoformat(timespec="seconds"),
            "created_at": ret.created_at.isoformat(timespec="seconds"),
        }
        await self._write_json(attempt_url / f"{pre}.{ret.id}.started.json", data)
        return ret

    async def finish_task(
        self,
        bake_id: str,
        attempt_no: int,
        task_no: int,
        cardinality: int,
        task: StartedTask,
        descr: JobDescription,
    ) -> FinishedTask:
        assert task.raw_id == descr.id
        assert task.created_at == descr.history.created_at
        bake_uri = self._mk_bake_uri(bake_id)
        attempt_url = bake_uri / f"{attempt_no:02d}.attempt"
        digits = cardinality // 10 + 1
        pre = str(task_no + 1).zfill(digits)
        assert descr.history.created_at is not None
        assert descr.history.started_at is not None
        assert descr.history.finished_at is not None
        ret = FinishedTask(
            id=task.id,
            raw_id=task.raw_id,
            when=datetime.datetime.now(datetime.timezone.utc),
            status=descr.history.status,
            exit_code=descr.history.exit_code,
            created_at=descr.history.created_at,
            started_at=descr.history.started_at,
            finished_at=descr.history.finished_at,
            finish_reason=descr.history.reason,
            finish_description=descr.history.description,
        )

        data = {
            "id": ret.id,
            "raw_id": ret.raw_id,
            "when": ret.when.isoformat(timespec="seconds"),
            "status": str(ret.status),
            "exit_code": ret.exit_code,
            "created_at": ret.created_at.isoformat(timespec="seconds"),
            "started_at": ret.started_at.isoformat(timespec="seconds"),
            "finished_at": ret.finished_at.isoformat(timespec="seconds"),
            "finish_reason": ret.finish_reason,
            "finish_description": ret.finish_description,
        }
        await self._write_json(attempt_url / f"{pre}.{task.id}.finished.json", data)
        return ret

    async def _read_file(self, url: URL) -> str:
        ret = []
        async for chunk in self._client.storage.open(url):
            ret.append(chunk)
        return b"".join(ret).decode("utf-8")

    async def _read_json(self, url: URL) -> Any:
        data = await self._read_file(url)
        return json.loads(data)

    async def _write_file(self, url: URL, body: str) -> None:
        # TODO: Prevent overriding the target on the storage.
        #
        # It might require platform_storage_api change.
        #
        # There is no clean understanding if the storage can support this strong
        # guarantee at all.
        files = set()
        async for fi in self._client.storage.ls(url.parent):
            files.add(fi.name)
        if url.name in files:
            raise ValueError(f"File {url} already exists")
        await self._client.storage.create(url, body.encode("utf-8"))

    async def _write_json(self, url: URL, data: Dict[str, Any]) -> None:
        if not data.get("when"):
            data["when"] = self._now()

        await self._write_file(url, json.dumps(data))



import dataclasses

import asyncio
import click
import datetime
import humanize
import shlex
import sys
from neuromation.api import Client, Factory, JobDescription, JobStatus, ResourceNotFound
from neuromation.cli.formatters import ftable  # TODO: extract into a separate library
from types import TracebackType
from typing import AbstractSet, List, Optional, Tuple, Type
from typing_extensions import AsyncContextManager

from . import ast
from .context import ImageCtx, LiveContext, UnknownJob, VolumeCtx


COLORS = {
    JobStatus.PENDING: "yellow",
    JobStatus.RUNNING: "blue",
    JobStatus.SUCCEEDED: "green",
    JobStatus.FAILED: "red",
    JobStatus.UNKNOWN: "yellow",
}


def format_job_status(status: JobStatus) -> str:
    return click.style(status.value, fg=COLORS.get(status, "reset"))


@dataclasses.dataclass(frozen=True)
class JobInfo:
    id: str
    status: JobStatus
    raw_id: Optional[str]  # low-level job id, None for never runned jobs
    tags: AbstractSet[str]
    when: Optional[datetime.datetime]


class LiveRunner(AsyncContextManager["LiveRunner"]):
    def __init__(self, flow: ast.LiveFlow) -> None:
        self._flow = flow
        self._ctx: Optional[LiveContext] = None
        self._client: Optional[Client] = None

    async def post_init(self) -> None:
        if self._ctx is not None:
            return
        self._ctx = await LiveContext.create(self._flow)
        self._client = await Factory().get()

    async def close(self) -> None:
        if self._client is not None:
            await self._client.close()

    async def __aenter__(self) -> "LiveRunner":
        await self.post_init()
        return self

    async def __aexit__(
        self,
        exc_typ: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    @property
    def ctx(self) -> LiveContext:
        assert self._ctx is not None
        return self._ctx

    @property
    def client(self) -> Client:
        assert self._client is not None
        return self._client

    async def run_subproc(self, exe: str, *args: str) -> None:
        proc = await asyncio.create_subprocess_exec(exe, *args)
        try:
            retcode = await proc.wait()
            if retcode:
                raise SystemExit(retcode)
        finally:
            if proc.returncode is None:
                # Kill neuro process if not finished
                # (e.g. if KeyboardInterrupt or cancellation was received)
                proc.kill()
                await proc.wait()

    async def ensure_job(self, job_id: str) -> LiveContext:
        try:
            return await self.ctx.with_job(job_id)
        except UnknownJob:
            click.secho(f"Unknown job {click.style(job_id, bold=True)}", fg="red")
            jobs = sorted([job for job in self._flow.jobs.keys()])
            jobs_str = ",".join(jobs)
            click.secho(f"Existing jobs: {jobs_str}", dim=True)
            sys.exit(1)

    async def resolve_job_by_name(
        self, name: Optional[str], tags: AbstractSet[str]
    ) -> JobDescription:
        async for job in self.client.jobs.list(
            name=name or "",
            tags=tags,
            reverse=True,
            limit=100,  # fixme: limit should be 1 but it doesn't work
        ):
            return job
        raise ResourceNotFound

    async def job_status(self, job_id: str) -> JobInfo:
        job_ctx = await self.ensure_job(job_id)
        job = job_ctx.job
        try:
            descr = await self.resolve_job_by_name(job.name, job.tags)
            if descr.status == JobStatus.PENDING:
                when = descr.history.created_at
            elif descr.status == JobStatus.RUNNING:
                when = descr.history.started_at
            else:
                when = descr.history.finished_at
            return JobInfo(job_id, descr.status, descr.id, job.tags, when)
        except ResourceNotFound:
            return JobInfo(job_id, JobStatus.UNKNOWN, None, job.tags, None)

    async def ps(self) -> None:
        """Return statuses for all jobs from the flow"""

        # TODO: make concurent queries for job statuses
        loop = asyncio.get_event_loop()
        rows: List[List[str]] = []
        rows.append(
            [
                click.style("JOB", bold=True),
                click.style("STATUS", bold=True),
                click.style("RAW ID", bold=True),
                click.style("WHEN", bold=True),
            ]
        )
        tasks = []
        for job_id in sorted(self._flow.jobs):
            tasks.append(loop.create_task(self.job_status(job_id)))

        for info in await asyncio.gather(*tasks):
            if info.when is None:
                when_humanized = "N/A"
            else:
                delta = datetime.datetime.now(datetime.timezone.utc) - info.when
                if delta < datetime.timedelta(days=1):
                    when_humanized = humanize.naturaltime(delta)
                else:
                    when_humanized = humanize.naturaldate(info.when.astimezone())
            rows.append(
                [
                    info.id,
                    format_job_status(info.status),
                    info.raw_id or "N/A",
                    when_humanized,
                ]
            )

        for line in ftable.table(rows):
            click.echo(line)

    async def status(self, job_id: str) -> None:
        job_ctx = await self.ensure_job(job_id)
        job = job_ctx.job
        try:
            descr = await self.resolve_job_by_name(job.name, job.tags)
            await self.run_subproc("neuro", "status", descr.id)
        except ResourceNotFound:
            click.echo(f"Job {click.style(job_id, bold=True)} is not running")
            sys.exit(1)

    async def run(self, job_id: str) -> None:
        """Run a named job"""
        job_ctx = await self.ensure_job(job_id)
        job = job_ctx.job

        try:
            descr = await self.resolve_job_by_name(job.name, job.tags)
            if descr.status == JobStatus.PENDING:
                click.echo(
                    f"Job {click.style(job_id, bold=True)} is pending, try again later"
                )
                sys.exit(2)
            if descr.status == JobStatus.RUNNING:
                click.echo(
                    f"Job {click.style(job_id, bold=True)} is running, connecting..."
                )
                # attach to job if needed, browse first
                if job.browse:
                    await self.run_subproc("neuro", "job", "browse", descr.id)
                if not job.detach:
                    await self.run_subproc("neuro", "attach", descr.id)
                return
            # Here the status is SUCCEDED or FAILED, restart
        except ResourceNotFound:
            # Job does not exist, run it
            pass

        args = ["run"]
        if job.title:
            args.append(f"--description={job.title}")
        if job.name:
            args.append(f"--name={job.name}")
        if job.preset is not None:
            args.append(f"--preset={job.preset}")
        if job.http_port is not None:
            args.append(f"--http={job.http_port}")
        if job.http_auth is not None:
            if job.http_auth:
                args.append(f"--http-auth")
            else:
                args.append(f"--no-http-auth")
        if job.entrypoint:
            args.append(f"--entrypoint={job.entrypoint}")
        if job.workdir is not None:
            raise NotImplementedError("workdir is not supported")
        for k, v in job_ctx.env.items():
            args.append(f"--env={k}={v}")
        for v in job.volumes:
            args.append(f"--volume={v}")
        for t in job.tags:
            args.append(f"--tag={t}")
        if job.life_span is not None:
            args.append(f"--life-span={int(job.life_span)}s")
        if job.browse:
            args.append(f"--browse")
        if job.detach:
            args.append(f"--detach")
        for pf in job.port_forward:
            args.append(f"--port-forward={pf}")

        args.append(job.image)
        if job.cmd:
            args.extend(shlex.split(job.cmd))
        await self.run_subproc("neuro", *args)

    async def logs(self, job_id: str) -> None:
        """Return job logs"""
        job_ctx = await self.ensure_job(job_id)
        job = job_ctx.job
        try:
            descr = await self.resolve_job_by_name(job.name, job.tags)
            await self.run_subproc("neuro", "logs", descr.id)
        except ResourceNotFound:
            click.echo(f"Job {click.style(job_id, bold=True)} is not running")
            sys.exit(1)

    async def kill_job(self, job_id: str) -> bool:
        """Kill named job"""
        job_ctx = await self.ensure_job(job_id)
        job = job_ctx.job

        try:
            descr = await self.resolve_job_by_name(job.name, job.tags)
            if descr.status in (JobStatus.PENDING, JobStatus.RUNNING):
                await self.client.jobs.kill(descr.id)
                descr = await self.client.jobs.status(descr.id)
                while descr.status not in (JobStatus.SUCCEEDED, JobStatus.FAILED):
                    await asyncio.sleep(0.2)
                    descr = await self.client.jobs.status(descr.id)
                return True
        except ResourceNotFound:
            pass
        return False

    async def kill(self, job_id: str) -> None:
        """Kill named job"""
        if await self.kill_job(job_id):
            click.echo(f"Killed job {click.style(job_id, bold=True)}")
        else:
            click.echo(f"Job {click.style(job_id, bold=True)} is not running")

    async def kill_all(self) -> None:
        """Kill all jobs"""
        tasks = []
        loop = asyncio.get_event_loop()

        async def kill(job_id: str) -> Tuple[str, bool]:
            return job_id, await self.kill_job(job_id)

        for job_id in sorted(self._flow.jobs):
            tasks.append(loop.create_task(kill(job_id)))

        for job_id, ret in await asyncio.gather(*tasks):
            if ret:
                click.echo(f"Killed job {click.style(job_id, bold=True)}")
            else:
                click.echo(f"Job {click.style(job_id, bold=True)} is not running")

    # volumes subsystem

    async def find_volume(self, volume: str) -> VolumeCtx:
        volume_ctx = self.ctx.volumes.get(volume)
        if volume_ctx is None:
            click.secho(f"Unknown volume {click.style(volume, bold=True)}", fg="red")
            volumes = sorted([volume for volume in self.ctx.volumes.keys()])
            volumes_str = ",".join(volumes)
            click.secho(f"Existing volumes: {volumes_str}", dim=True)
            sys.exit(1)
        if volume_ctx.local is None:
            click.echo(f"Volume's {click.style('local', bold=True)} part is not set")
            sys.exit(2)
        return volume_ctx

    async def upload(self, volume: str) -> None:
        volume_ctx = await self.find_volume(volume)
        await self.run_subproc(
            "neuro",
            "cp",
            "--recursive",
            "--update",
            "--no-target-directory",
            str(volume_ctx.full_local_path),
            str(volume_ctx.remote),
        )

    async def download(self, volume: str) -> None:
        volume_ctx = await self.find_volume(volume)
        await self.run_subproc(
            "neuro" "cp",
            "--recursive",
            "--update",
            "--no-target-directory",
            str(volume_ctx.remote),
            str(volume_ctx.full_local_path),
        )

    async def clean(self, volume: str) -> None:
        volume_ctx = await self.find_volume(volume)
        await self.run_subproc("neuro", "rm", "--recursive", str(volume_ctx.remote))

    async def upload_all(self) -> None:
        for volume in self.ctx.volumes.values():
            if volume.local is not None:
                await self.upload(volume.id)

    async def download_all(self) -> None:
        for volume in self.ctx.volumes.values():
            if volume.local is not None:
                await self.download(volume.id)

    async def clean_all(self) -> None:
        for volume in self.ctx.volumes.values():
            if volume.local is not None:
                await self.clean(volume.id)

    async def mkvolumes(self) -> None:
        for volume in self.ctx.volumes.values():
            if volume.local is not None:
                volume_ctx = await self.find_volume(volume.id)
                click.echo(f"Create volume {click.style(volume.id, bold=True)}")
                await self.run_subproc(
                    "neuro", "mkdir", "--parents", str(volume_ctx.remote),
                )

    # images subsystem

    async def find_image(self, image: str) -> ImageCtx:
        image_ctx = self.ctx.images.get(image)
        if image_ctx is None:
            click.secho(f"Unknown image {click.style(image, bold=True)}", fg="red")
            images = sorted([image for image in self.ctx.images.keys()])
            images_str = ",".join(images)
            click.secho(f"Existing images: {images_str}", dim=True)
            sys.exit(1)
        if image_ctx.context is None:
            click.echo(f"Image's {click.style('context', bold=True)} part is not set")
            sys.exit(2)
        if image_ctx.dockerfile is None:
            click.echo(
                f"Image's {click.style('dockerfile', bold=True)} part is not set"
            )
            sys.exit(3)
        return image_ctx

    async def build(self, image: str) -> None:
        image_ctx = await self.find_image(image)
        cmd = []
        assert image_ctx.full_dockerfile_path is not None
        assert image_ctx.full_context_path is not None
        rel_dockerfile_path = image_ctx.full_dockerfile_path.relative_to(
            image_ctx.full_context_path
        )
        cmd.append(f"--file={rel_dockerfile_path}")
        for arg in image_ctx.build_args:
            cmd.append(f"--build-arg=arg")
        cmd.append(str(image_ctx.full_context_path))
        cmd.append(str(image_ctx.ref))
        await self.run_subproc("neuro-extras", "image", "build", *cmd)



import asyncio
import sys
import tempfile
from logging import getLogger
from neuromation.api import (
    Client,
    Container,
    HTTPPort,
    JobDescription,
    JobStatus,
    Resources,
    SecretFile,
    Volume,
)
from neuromation.api.url_utils import uri_from_cli
from types import TracebackType
from typing import Dict, Iterable, Optional, Set, Type
from typing_extensions import AsyncContextManager
from yarl import URL

from . import ast
from .commands import CmdProcessor
from .context import BatchContext, DepCtx, Result
from .parser import ConfigDir, parse_batch
from .storage import BatchStorage, FinishedTask, StartedTask
from .types import LocalPath


if sys.version_info >= (3, 9):
    import graphlib
else:
    from . import backport_graphlib as graphlib


log = getLogger(__name__)


class BatchRunner(AsyncContextManager["BatchRunner"]):
    def __init__(
        self, config_dir: ConfigDir, client: Client, storage: BatchStorage
    ) -> None:
        self._config_dir = config_dir
        self._client = client
        self._storage = storage

    async def close(self) -> None:
        pass

    async def __aenter__(self) -> "BatchRunner":
        return self

    async def __aexit__(
        self,
        exc_typ: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def run_subproc(self, exe: str, *args: str) -> None:
        proc = await asyncio.create_subprocess_exec(exe, *args)
        try:
            retcode = await proc.wait()
            if retcode:
                raise SystemExit(retcode)
        finally:
            if proc.returncode is None:
                # Kill neuro process if not finished
                # (e.g. if KeyboardInterrupt or cancellation was received)
                proc.kill()
                await proc.wait()

    async def bake(self, batch_name: str) -> None:
        # batch_name is a name of yaml config inside self._workspace / .neuro
        # folder without the file extension
        config_file = (self._config_dir.config_dir / (batch_name + ".yml")).resolve()

        log.info("Use config file %s", config_file)

        log.info("Check config")
        # Check that the yaml is parseable
        flow = parse_batch(self._config_dir.workspace, config_file)
        assert isinstance(flow, ast.BatchFlow)

        ctx = await BatchContext.create(flow)
        for volume in ctx.volumes.values():
            if volume.local is not None:
                # TODO: sync volumes if needed
                raise NotImplementedError("Volumes sync is not supported")

        toposorter = graphlib.TopologicalSorter(ctx.graph)
        # check fast for the graph cycle error
        toposorter.prepare()

        log.info("Config is correct")

        config_content = config_file.read_text()

        log.info("Create bake")
        bake_id = await self._storage.create_bake(
            batch_name, config_file.name, config_content, ctx.cardinality
        )
        log.info("Bake %s created", bake_id)
        await self._storage.create_attempt(bake_id, 1, ctx.cardinality)
        log.info("Start attempt %d", 1)
        # TODO: run this function in a job
        await self.process(bake_id)

    async def process(self, bake_id: str) -> None:
        log.info("Process %s", bake_id)
        with tempfile.TemporaryDirectory(prefix=bake_id) as tmp:
            root_dir = LocalPath(tmp)
            log.info("Root dir %s", root_dir)
            config_dir = root_dir / ".neuro"
            config_dir.mkdir()
            workspace = config_dir / "workspace"
            workspace.mkdir()

            log.info("Fetch bake init")
            bake_init = await self._storage.fetch_bake(bake_id)
            log.info("Fetch baked config")
            config_content = await self._storage.fetch_config(
                bake_id, bake_init.config_file.name
            )
            config_file = config_dir / bake_init.config_file.name
            config_file.write_text(config_content)

            log.info("Parse baked config")
            flow = parse_batch(workspace, config_file)
            assert isinstance(flow, ast.BatchFlow)

            ctx = await BatchContext.create(flow)

            log.info("Find last attempt")
            attempt = await self._storage.find_last_attempt(bake_id)
            log.info("Fetch attempt %d", attempt)
            finished, started = await self._storage.fetch_attempt(
                bake_id, attempt, ctx.cardinality
            )

            toposorter = graphlib.TopologicalSorter(ctx.graph)
            toposorter.prepare()
            for tid in finished:
                toposorter.done(tid)

            while True:
                for tid in toposorter.get_ready():
                    if tid in started:
                        continue
                    deps = ctx.graph[tid]
                    needs = {}
                    for dep_id in deps:
                        dep = finished.get(dep_id)
                        assert dep is not None
                        needs[dep_id] = DepCtx(dep.result, {})
                    task_ctx = await ctx.with_task(tid, needs=needs)
                    log.info("Task %s started", tid)
                    st = await self._start_task(
                        bake_id, attempt, len(started) + len(finished), task_ctx
                    )
                    started[st.id] = st

                for st in started.values():
                    if st.id in finished:
                        continue
                    status = await self._client.jobs.status(st.raw_id)
                    if status.status in (JobStatus.FAILED, JobStatus.SUCCEEDED):
                        log.info("Task %s finished", tid)
                        finished[st.id] = await self._finish_task(
                            bake_id,
                            attempt,
                            len(started) + len(finished),
                            ctx.cardinality,
                            st,
                            status,
                        )
                        toposorter.done(st.id)

                if len(finished) == ctx.cardinality // 2:
                    log.info("Attempt %d finished", attempt)
                    await self._storage.finish_attempt(
                        bake_id,
                        attempt,
                        ctx.cardinality,
                        self._accumulate_result(finished.values()),
                    )
                    return

                # AS: I have no idea what timeout is better;
                # too short value bombards servers,
                # too long timeout makes the waiting longer than expected
                # The missing events subsystem would be great for this task :)
                await asyncio.sleep(1)

    def _accumulate_result(self, finished: Iterable[FinishedTask]) -> Result:
        # TODO: handle cancelled tasks
        for task in finished:
            if task.status == JobStatus.FAILED:
                return Result.FAILED

        return Result.SUCCEEDED

    async def _start_task(
        self, bake_id: str, attempt: int, task_no: int, task_ctx: BatchContext
    ) -> StartedTask:
        task = task_ctx.task

        preset_name = task.preset
        if preset_name is None:
            preset_name = next(iter(self._client.config.presets))
        preset = self._client.config.presets[preset_name]

        env_dict = dict(task_ctx.env)
        secret_env_dict = self._extract_secret_env(env_dict)
        resources = Resources(
            memory_mb=preset.memory_mb,
            cpu=preset.cpu,
            gpu=preset.gpu,
            gpu_model=preset.gpu_model,
            shm=True,
            tpu_type=preset.tpu_type,
            tpu_software_version=preset.tpu_software_version,
        )
        input_secret_files = {vol for vol in task.volumes if vol.startswith("secret:")}
        input_volumes = set(task.volumes) - input_secret_files
        secret_files = await self._build_secret_files(input_secret_files)
        volumes = await self._build_volumes(input_volumes)

        http_auth = task.http_auth
        if http_auth is None:
            http_auth = HTTPPort.requires_auth

        container = Container(
            image=self._client.parse.remote_image(task.image),
            entrypoint=task.entrypoint,
            command=task.cmd,
            http=HTTPPort(task.http_port, http_auth) if task.http_port else None,
            resources=resources,
            env=env_dict,
            volumes=list(volumes),
            secret_env=secret_env_dict,
            secret_files=list(secret_files),
            tty=False,
        )

        job = await self._client.jobs.run(
            container,
            is_preemptible=preset.is_preemptible,
            name=task.name,
            tags=list(task.tags),
            description=task.title,
            life_span=task.life_span,
        )
        return await self._storage.start_task(
            bake_id, attempt, task_no, task_ctx.cardinality, task.real_id, job
        )

    def _extract_secret_env(self, env_dict: Dict[str, str]) -> Dict[str, URL]:
        secret_env_dict = {}
        for name, val in env_dict.copy().items():
            if val.startswith("secret:"):
                secret_env_dict[name] = uri_from_cli(
                    val,
                    self._client.username,
                    self._client.cluster_name,
                    allowed_schemes=("secret"),
                )
                del env_dict[name]
        return secret_env_dict

    async def _build_volumes(self, input_volumes: Iterable[str]) -> Set[Volume]:
        volumes: Set[Volume] = set()

        for vol in input_volumes:
            volumes.add(self._client.parse.volume(vol))
        return volumes

    async def _build_secret_files(self, input_volumes: Set[str]) -> Set[SecretFile]:
        secret_files: Set[SecretFile] = set()
        for volume in input_volumes:
            parts = volume.split(":")
            if len(parts) != 3:
                raise ValueError(f"Invalid secret file specification '{volume}'")
            container_path = parts.pop()
            secret_uri = uri_from_cli(
                ":".join(parts),
                self._client.username,
                self._client.cluster_name,
                allowed_schemes=("secret"),
            )
            secret_files.add(SecretFile(secret_uri, container_path))
        return secret_files

    async def _finish_task(
        self,
        bake_id: str,
        attempt_no: int,
        task_no: int,
        cardinality: int,
        task: StartedTask,
        descr: JobDescription,
    ) -> FinishedTask:
        async with CmdProcessor() as proc:
            buf = bytearray()
            async for chunk in self._client.jobs.monitor(task.raw_id):
                buf.extend(chunk)
                if b"\n" in buf:
                    blines = buf.splitlines(keepends=True)
                    buf = blines.pop(-1)
                    for bline in blines:
                        line = bline.decode("utf-8", "replace")
                        await proc.feed(line)
            line = buf.decode("utf-8", "replace")
            await proc.feed(line)
        return await self._storage.finish_task(
            bake_id, attempt_no, task_no, cardinality, task, descr
        )



# Workflow commands that can be used in job's output
# Candidates:
# ::set-env::
# ::set-output::
# ::save-state::

from typing import AsyncContextManager


class CmdProcessor(AsyncContextManager["CmdProcessor"]):
    def __init__(self) -> None:
        pass

    async def feed(self, line: str) -> None:
        pass



# YAML parser
#
# The parser converts YAML entities into ast data classes.
#
# Defaults are evaluated by the separate processing step.


import dataclasses

import abc
import yaml
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    TextIO,
    Type,
    TypeVar,
    Union,
)
from yaml.composer import Composer
from yaml.constructor import ConstructorError, SafeConstructor
from yaml.parser import Parser
from yaml.reader import Reader
from yaml.resolver import Resolver
from yaml.scanner import Scanner

from . import ast
from .expr import (
    Expr,
    IdExpr,
    OptBashExpr,
    OptBoolExpr,
    OptIdExpr,
    OptIntExpr,
    OptLifeSpanExpr,
    OptLocalPathExpr,
    OptPythonExpr,
    OptRemotePathExpr,
    OptStrExpr,
    PortPairExpr,
    RemotePathExpr,
    StrExpr,
    URIExpr,
)
from .tokenizer import Pos
from .types import LocalPath


_T = TypeVar("_T")
_Cont = TypeVar("_Cont")


@dataclasses.dataclass
class ConfigDir:
    workspace: LocalPath
    config_dir: LocalPath


@dataclasses.dataclass
class ConfigPath:
    workspace: LocalPath
    config_file: LocalPath


class ConfigConstructor(SafeConstructor):
    def __init__(self, id: str, workspace: LocalPath) -> None:
        super().__init__()
        self._id = id
        self._workspace = workspace

    def construct_id(self, node: yaml.Node) -> str:
        val = self.construct_object(node)  # type: ignore[no-untyped-call]
        if not isinstance(val, str):
            raise ConstructorError(
                None, None, f"expected a str, found {type(val)}", node.start_mark
            )
        if not val.isidentifier():
            raise ConstructorError(
                None, None, f"{val} is not an identifier", node.start_mark
            )
        if val == val.upper():
            raise ConstructorError(
                None,
                None,
                f"{val} is invalid identifier, "
                "uppercase names are reserved for internal usage",
                node.start_mark,
            )
        return val


class Loader(Reader, Scanner, Parser, Composer, ConfigConstructor, Resolver):
    def __init__(self, stream: TextIO, id: str, workspace: LocalPath) -> None:
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        ConfigConstructor.__init__(self, id, workspace)
        Resolver.__init__(self)


def mark2pos(mark: yaml.Mark) -> Pos:
    return Pos(mark.line, mark.column, LocalPath(mark.name))


class SimpleCompound(Generic[_T, _Cont], abc.ABC):
    def __init__(self, factory: Type[Expr[_T]]) -> None:
        self._factory = factory

    @abc.abstractmethod
    def construct(self, ctor: ConfigConstructor, node: yaml.Node) -> _Cont:
        pass

    def check_scalar(self, ctor: ConfigConstructor, node: yaml.Node) -> None:
        node_id = node.id  # type: ignore
        if node_id != "scalar":
            raise ConstructorError(
                None,
                None,
                f"expected a scalar node, but found {node_id}",
                node.start_mark,
            )


class SimpleSeq(SimpleCompound[_T, Sequence[Expr[_T]]]):
    def construct(self, ctor: ConfigConstructor, node: yaml.Node) -> Sequence[Expr[_T]]:
        if not isinstance(node, yaml.SequenceNode):
            node_id = node.id  # type: ignore
            raise ConstructorError(
                None,
                None,
                f"expected a sequence node, but found {node_id}",
                node.start_mark,
            )
        ret = []
        for child in node.value:
            self.check_scalar(ctor, child)
            val = ctor.construct_object(child)  # type: ignore[no-untyped-call]
            # check if scalar
            if val is not None:
                val = str(val)
            tmp = self._factory(
                mark2pos(child.start_mark), mark2pos(child.end_mark), val
            )
            ret.append(tmp)
        return ret


class SimpleMapping(SimpleCompound[_T, Mapping[str, Expr[_T]]]):
    def construct(
        self, ctor: ConfigConstructor, node: yaml.Node
    ) -> Mapping[str, Expr[_T]]:
        if not isinstance(node, yaml.MappingNode):
            node_id = node.id  # type: ignore
            raise ConstructorError(
                None,
                None,
                f"expected a mapping node, but found {node_id}",
                node.start_mark,
            )
        ret = {}
        for k, v in node.value:
            key = ctor.construct_object(k)  # type: ignore[no-untyped-call]
            self.check_scalar(ctor, v)
            tmp = ctor.construct_object(v)  # type: ignore[no-untyped-call]
            if tmp is not None:
                tmp = str(tmp)
            value = self._factory(mark2pos(v.start_mark), mark2pos(v.end_mark), tmp)
            ret[key] = value
        return ret


class IdMapping(SimpleCompound[_T, Mapping[str, Expr[_T]]]):
    def construct(
        self, ctor: ConfigConstructor, node: yaml.Node
    ) -> Mapping[str, Expr[_T]]:
        if not isinstance(node, yaml.MappingNode):
            node_id = node.id  # type: ignore
            raise ConstructorError(
                None,
                None,
                f"expected a mapping node, but found {node_id}",
                node.start_mark,
            )
        ret = {}
        for k, v in node.value:
            key = ctor.construct_id(k)
            self.check_scalar(ctor, v)
            tmp = ctor.construct_object(v)  # type: ignore[no-untyped-call]
            if tmp is not None:
                tmp = str(tmp)
            value = self._factory(mark2pos(v.start_mark), mark2pos(v.end_mark), tmp)
            ret[key] = value
        return ret


KeyT = Union[
    None, Type[Expr[Any]], SimpleCompound[Any, Any], Type[ast.Kind],
]
VarT = Union[
    None, Expr[Any], Mapping[str, Any], Sequence[Any], ast.Base, ast.Kind,
]

_AstType = TypeVar("_AstType", bound=ast.Base)


def parse_dict(
    ctor: ConfigConstructor,
    node: yaml.MappingNode,
    keys: Mapping[str, KeyT],
    res_type: Type[_AstType],
    *,
    ret_name: Optional[str] = None,
    extra: Optional[Mapping[str, Union[str, LocalPath]]] = None,
    preprocess: Optional[
        Callable[[ConfigConstructor, Dict[str, VarT]], Dict[str, VarT]]
    ] = None,
    find_res_type: Optional[
        Callable[[ConfigConstructor, Type[_AstType], Dict[str, VarT]], Type[_AstType]]
    ] = None,
) -> _AstType:
    if extra is None:
        extra = {}
    if ret_name is None:
        ret_name = res_type.__name__
    if not isinstance(node, yaml.MappingNode):
        node_id = node.id
        raise ConstructorError(
            None,
            None,
            f"expected a mapping node, but found {node_id}",
            node.start_mark,
        )
    node_start = mark2pos(node.start_mark)
    node_end = mark2pos(node.end_mark)

    data = {}
    for k, v in node.value:
        key = ctor.construct_object(k)  # type: ignore[no-untyped-call]
        if key not in keys:
            raise ConstructorError(
                f"while constructing a {ret_name}",
                node.start_mark,
                f"unexpected key {key}",
                k.start_mark,
            )
        item_ctor: KeyT = keys[key]
        tmp: Any
        value: VarT
        if item_ctor is None:
            # Get constructor from tag
            value = ctor.construct_object(v)  # type: ignore[no-untyped-call]
        elif item_ctor is ast.Kind:
            tmp = str(ctor.construct_object(v))  # type: ignore[no-untyped-call]
            value = ast.Kind(tmp)
        elif isinstance(item_ctor, ast.Base):
            assert isinstance(
                v, ast.Base
            ), f"[{type(v)}] {v} should be ast.Base derived"
            value = v
        elif isinstance(item_ctor, SimpleCompound):
            value = item_ctor.construct(ctor, v)
        elif isinstance(item_ctor, type) and issubclass(item_ctor, Expr):
            tmp = str(ctor.construct_object(v))  # type: ignore[no-untyped-call]
            value = item_ctor(mark2pos(v.start_mark), mark2pos(v.end_mark), tmp)
        else:
            raise ConstructorError(
                f"while constructing a {ret_name}",
                node.start_mark,
                f"unexpected value tag {v.tag} for key {key}[{item_ctor}]",
                k.start_mark,
            )
        data[key] = value

    if preprocess is not None:
        data = preprocess(ctor, dict(data))
    if find_res_type is not None:
        res_type = find_res_type(ctor, res_type, dict(data))
        ret_name = res_type.__name__

    optional_fields: Dict[str, Any] = {}
    found_fields = extra.keys() | data.keys() | {"_start", "_end"}
    for f in dataclasses.fields(res_type):
        if f.name not in found_fields:
            key = f.name
            item_ctor = keys[key]
            if item_ctor is None:
                optional_fields[f.name] = None
            elif isinstance(item_ctor, SimpleCompound):
                optional_fields[f.name] = None
            elif isinstance(item_ctor, ast.Base):
                optional_fields[f.name] = None
            elif isinstance(item_ctor, type) and issubclass(item_ctor, Expr):
                if not item_ctor.allow_none:
                    raise ConstructorError(
                        f"while constructing a {ret_name}, "
                        f"missing mandatory key {f.name}",
                        node.start_mark,
                    )
                optional_fields[f.name] = item_ctor(node_start, node_end, None)
            else:
                raise ConstructorError(
                    f"while constructing a {ret_name}, "
                    f"unexpected {f.name} constructor type {item_ctor!r}",
                    node.start_mark,
                )
    return res_type(  # type: ignore[call-arg]
        _start=node_start, _end=node_end, **extra, **data, **optional_fields,
    )


def parse_volume(ctor: ConfigConstructor, node: yaml.MappingNode) -> ast.Volume:
    return parse_dict(
        ctor,
        node,
        {
            "remote": URIExpr,
            "mount": RemotePathExpr,
            "read_only": OptBoolExpr,
            "local": OptLocalPathExpr,
        },
        ast.Volume,
    )


def parse_volumes(
    ctor: ConfigConstructor, node: yaml.MappingNode
) -> Dict[str, ast.Volume]:
    ret = {}
    for k, v in node.value:
        key = ctor.construct_id(k)
        value = parse_volume(ctor, v)
        ret[key] = value
    return ret


Loader.add_path_resolver("flow:volumes", [(dict, "volumes")])  # type: ignore
Loader.add_constructor("flow:volumes", parse_volumes)  # type: ignore


def parse_image(ctor: ConfigConstructor, node: yaml.MappingNode) -> ast.Image:
    return parse_dict(
        ctor,
        node,
        {
            "ref": StrExpr,
            "context": OptLocalPathExpr,
            "dockerfile": OptLocalPathExpr,
            "build_args": SimpleSeq(StrExpr),
        },
        ast.Image,
    )


def parse_images(
    ctor: ConfigConstructor, node: yaml.MappingNode
) -> Dict[str, ast.Image]:
    ret = {}
    for k, v in node.value:
        key = ctor.construct_id(k)
        value = parse_image(ctor, v)
        ret[key] = value
    return ret


Loader.add_path_resolver("flow:images", [(dict, "images")])  # type: ignore
Loader.add_constructor("flow:images", parse_images)  # type: ignore


def parse_exc_inc(
    ctor: ConfigConstructor, node: yaml.MappingNode
) -> Sequence[Mapping[str, StrExpr]]:
    if not isinstance(node, yaml.SequenceNode):
        node_id = node.id
        raise ConstructorError(
            None,
            None,
            f"expected a sequence node, but found {node_id}",
            node.start_mark,
        )
    builder = IdMapping(StrExpr)
    ret: List[Mapping[str, StrExpr]] = []
    for v in node.value:
        ret.append(builder.construct(ctor, v))  # type: ignore[arg-type]
    return ret


def parse_matrix(ctor: ConfigConstructor, node: yaml.MappingNode) -> ast.Matrix:
    if not isinstance(node, yaml.MappingNode):
        node_id = node.id
        raise ConstructorError(
            None,
            None,
            f"expected a mapping node, but found {node_id}",
            node.start_mark,
        )
    products_builder = SimpleSeq(StrExpr)
    products = {}
    exclude: Sequence[Mapping[str, StrExpr]] = []
    include: Sequence[Mapping[str, StrExpr]] = []
    for k, v in node.value:
        key = ctor.construct_id(k)
        if key == "include":
            include = parse_exc_inc(ctor, v)
        elif key == "exclude":
            exclude = parse_exc_inc(ctor, v)
        else:
            products[key] = products_builder.construct(ctor, v)
    return ast.Matrix(
        _start=mark2pos(node.start_mark),
        _end=mark2pos(node.end_mark),
        products=products,  # type: ignore[arg-type]
        exclude=exclude,
        include=include,
    )


Loader.add_path_resolver(  # type: ignore
    "flow:matrix",
    [(dict, "tasks"), (list, None), (dict, "strategy"), (dict, "matrix")],
)
Loader.add_constructor("flow:matrix", parse_matrix)  # type: ignore


STRATEGY = {
    "matrix": None,
    "fail_fast": OptBoolExpr,
    "max_parallel": OptIntExpr,
}


def parse_strategy(ctor: ConfigConstructor, node: yaml.MappingNode) -> ast.Strategy:
    return parse_dict(
        ctor,
        node,
        {"matrix": None, "fail_fast": OptBoolExpr, "max_parallel": OptIntExpr},
        ast.Strategy,
    )


Loader.add_path_resolver(  # type: ignore
    "flow:strategy", [(dict, "tasks"), (list, None), (dict, "strategy")]
)
Loader.add_constructor("flow:strategy", parse_strategy)  # type: ignore


EXEC_UNIT = {
    "title": OptStrExpr,
    "name": OptStrExpr,
    "image": StrExpr,
    "preset": OptStrExpr,
    "entrypoint": OptStrExpr,
    "cmd": OptStrExpr,
    "bash": OptBashExpr,
    "python": OptPythonExpr,
    "workdir": OptRemotePathExpr,
    "env": SimpleMapping(StrExpr),
    "volumes": SimpleSeq(StrExpr),
    "tags": SimpleSeq(StrExpr),
    "life_span": OptLifeSpanExpr,
    "http_port": OptIntExpr,
    "http_auth": OptBoolExpr,
}

JOB = {
    "detach": OptBoolExpr,
    "browse": OptBoolExpr,
    "port_forward": SimpleSeq(PortPairExpr),
    **EXEC_UNIT,
}


def select_shells(ctor: ConfigConstructor, dct: Dict[str, Any]) -> Dict[str, Any]:
    found = {k for k in dct if k in ("cmd", "bash", "python")}
    if len(found) > 1:
        raise ValueError(f"{','.join(found)} are mutually exclusive")

    bash = dct.pop("bash", None)
    if bash is not None:
        dct["cmd"] = bash

    python = dct.pop("python", None)
    if python is not None:
        dct["cmd"] = python

    return dct


def parse_job(ctor: ConfigConstructor, node: yaml.MappingNode) -> ast.Job:
    return parse_dict(
        ctor, node, JOB, ast.Job, preprocess=select_shells  # type: ignore[arg-type]
    )


def parse_jobs(ctor: ConfigConstructor, node: yaml.MappingNode) -> Dict[str, ast.Job]:
    ret = {}
    for k, v in node.value:
        key = ctor.construct_id(k)
        value = parse_job(ctor, v)
        ret[key] = value
    return ret


Loader.add_path_resolver("flow:jobs", [(dict, "jobs")])  # type: ignore
Loader.add_constructor("flow:jobs", parse_jobs)  # type: ignore


TASK = {
    "id": OptIdExpr,
    "needs": SimpleSeq(IdExpr),
    "strategy": None,
    **EXEC_UNIT,
}


def parse_task(ctor: ConfigConstructor, node: yaml.MappingNode) -> ast.Task:
    return parse_dict(
        ctor, node, TASK, ast.Task, preprocess=select_shells  # type: ignore
    )


Loader.add_path_resolver(  # type: ignore[no-untyped-call]
    "flow:task", [(dict, "tasks"), (list, None)]
)
Loader.add_constructor("flow:task", parse_task)  # type: ignore


Loader.add_path_resolver("flow:tasks", [(dict, "tasks")])  # type: ignore
Loader.add_constructor("flow:tasks", Loader.construct_sequence)  # type: ignore


def parse_flow_defaults(
    ctor: ConfigConstructor, node: yaml.MappingNode
) -> ast.FlowDefaults:
    return parse_dict(
        ctor,
        node,
        {
            "tags": SimpleSeq(StrExpr),
            "env": SimpleMapping(StrExpr),
            "workdir": OptRemotePathExpr,
            "life_span": OptLifeSpanExpr,
            "preset": OptStrExpr,
        },
        ast.FlowDefaults,
    )


Loader.add_path_resolver("flow:defaults", [(dict, "defaults")])  # type: ignore
Loader.add_constructor("flow:defaults", parse_flow_defaults)  # type: ignore


FLOW = {
    "kind": ast.Kind,
    "id": None,
    "title": None,
    "images": None,
    "volumes": None,
    "defaults": None,
    "jobs": None,
    "tasks": None,
}


def parse_opt_str(ctor: ConfigConstructor, node: yaml.MappingNode) -> str:
    return str(ctor.construct_scalar(node))  # type: ignore[no-untyped-call]


Loader.add_path_resolver("flow:opt_str", [(dict, "title")])  # type: ignore
Loader.add_constructor("flow:opt_str", parse_opt_str)  # type: ignore


def select_kind(ctor: ConfigConstructor, dct: Dict[str, Any]) -> Dict[str, Any]:
    if dct["kind"] == ast.Kind.LIVE:
        tasks = dct.pop("tass", None)
        if tasks is not None:
            raise ValueError("flow of kind={dct['kind']} cannot have tasks")
    elif dct["kind"] == ast.Kind.BATCH:
        jobs = dct.pop("jobs", None)
        if jobs is not None:
            raise ValueError("flow of kind={dct['kind']} cannot have jobs")
        del dct["jobs"]
    else:
        raise ValueError(f"Unknown kind {dct['kind']} of the flow")
    return dct


def find_res_type(
    ctor: ConfigConstructor, res_type: Type[ast.BaseFlow], arg: Dict[str, VarT]
) -> Type[ast.BaseFlow]:
    if arg["kind"] == ast.Kind.LIVE:
        return ast.LiveFlow
    elif arg["kind"] == ast.Kind.BATCH:
        return ast.BatchFlow
    else:
        raise ValueError(f"Unknown kind {arg['kind']} of the flow")


def parse_main(ctor: ConfigConstructor, node: yaml.MappingNode) -> ast.BaseFlow:
    return parse_dict(
        ctor,
        node,
        FLOW,
        ast.BaseFlow,
        find_res_type=find_res_type,
        extra={"id": ctor._id, "workspace": ctor._workspace},
    )


Loader.add_path_resolver("flow:main", [])  # type: ignore
Loader.add_constructor("flow:main", parse_main)  # type: ignore


def parse_live(
    workspace: Path, config_file: Path, *, id: Optional[str] = None
) -> ast.LiveFlow:
    # Parse live flow config file
    if id is None:
        id = workspace.stem
    with config_file.open() as f:
        loader = Loader(f, id, workspace)
        try:
            ret = loader.get_single_data()  # type: ignore[no-untyped-call]
            assert isinstance(ret, ast.LiveFlow)
            assert ret.kind == ast.Kind.LIVE
            return ret
        finally:
            loader.dispose()  # type: ignore[no-untyped-call]


def parse_batch(
    workspace: Path, config_file: Path, *, id: Optional[str] = None
) -> ast.BatchFlow:
    # Parse pipeline flow config file
    if id is None:
        id = config_file.stem
    with config_file.open() as f:
        loader = Loader(f, id, workspace)
        try:
            ret = loader.get_single_data()  # type: ignore[no-untyped-call]
            assert isinstance(ret, ast.BatchFlow)
            assert ret.kind == ast.Kind.BATCH
            return ret
        finally:
            loader.dispose()  # type: ignore[no-untyped-call]


def find_workspace(path: Optional[Union[Path, str]]) -> ConfigDir:
    if path is not None:
        if not isinstance(path, Path):
            path = Path(path)
        if not path.exists():
            raise ValueError(f"{path} does not exist")
        if not path.is_dir():
            raise ValueError(f"{path} should be a directory")
    else:
        path = Path.cwd()

    orig_path = path

    while True:
        if path == path.parent:
            raise ValueError(f".neuro folder was not found in lookup for {orig_path}")
        if (path / ".neuro").is_dir():
            break
        path = path.parent

    return ConfigDir(path, path / ".neuro")


def find_live_config(workspace: ConfigDir) -> ConfigPath:
    # Find live config file, starting from path.
    # Return a project root folder and a path to config file.
    #
    # If path is a file -- it is used as is.
    # If path is a directory -- it is used as starting point, Path.cwd() otherwise.
    # The lookup searches bottom-top from path dir up to the root folder,
    # looking for .neuro folder and ./neuro/live.yml
    # If the config file not found -- raise an exception.

    ret = workspace.config_dir / "live.yml"
    if not ret.exists():
        raise ValueError(f"{ret} does not exist")
    if not ret.is_file():
        raise ValueError(f"{ret} is not a file")
    return ConfigPath(workspace.workspace, ret)
